//#define DEBUG


#include <msp430x14x1.h>
#include <io.h>
#include <signal.h>
#include <adc12.h>

#ifdef DEBUG
#include <stdio.h>
#endif

typedef unsigned long       u32;
typedef signed long         s32;
typedef unsigned int        u16;
typedef signed int          s16;
typedef signed char         s8;
typedef unsigned char       u8;

#define B0	((u8)0x01)
#define B1	((u8)0x02)
#define B2	((u8)0x04)
#define B3	((u8)0x08)
#define B4	((u8)0x10)
#define B5	((u8)0x20)
#define B6	((u8)0x40)
#define B7	((u8)0x80)

// pressure sensor sampling states
#define SENSOR_READ_TEMPERATURE 		11
#define SENSOR_READ_PRESSURE			22

#define FOSC_HZ             270000L

#define ABS(x)                 ((x) < 0 ? -(x) : (x))
#define CLAMP(x,min,max)       {if ((x) <= (min)) (x) = (min); else if ((x) >= (max)) (x) = (max);}
#define CORE(x,t)              {if (ABS(x) <= (t)) (x) = 0;}

// ~3 second z buffer

#define SENSOR_SAMPLES_PER_SEC  11L
#define SNS_NUM_Z_SAMPLES       35L  

#define BEEP_DISCR_THRESHOLD    30L

#define VARIO_MAX_CPS         1000L
#define VARIO_XOVER_CPS       400L

#define VARIO_MAX_FREQHZ      5000L
#define VARIO_XOVER_FREQHZ    2000L
#define VARIO_MIN_FREQHZ       400L

#define VARIO_SINK_FREQHZ     1000L

#define SINKTONE_CPS          -300L
#define CLIMBTONE_CPS           10L

#define TMRB_STOP    (TBSSEL_2 + ID_3 + MC_0)
#define TMRB_START   (TBSSEL_2 + ID_3 + MC_2)

// MCLK P4.2
// DIN  P4.1
// DOUT P4.0
// SCLK P3.7


#define DIN_0()      { P4OUT &= ~B1;}
#define DIN_1()      { P4OUT |=  B1;}

#define SCLK_0()     { P3OUT &= ~B7;}
#define SCLK_1()     { P3OUT |= B7;}

#define DOUT         ((u16)(( P4IN & B0 ) ? 1 : 0 ))


// OPT_ACCEL  P1.5
// OPT_CHIRP  P1.6

// uncomment the #defines if you want jumper/DIP switch enabled options. 
// To enable, connect port pin to VCC via 100k resistor.
// To disable, ground the port pin

// #define OPT_ACCEL   (P1IN & B5)
// #define OPT_CHIRP   (P1IN & B6)

// if you uncomment the above two #defines, make sure you 
// comment out the next two :-)

#define OPT_ACCEL   0
#define OPT_CHIRP   1

#define KLPF_ACCEL   		25

u16 gwD1;
u16 gwD2;
u16 gwC1;
u16 gwC2;
u16 gwC3;
u16 gwC4;
u16 gwC5;
u16 gwC6;
   
s32 gP;
s16 gT;

s16 gnTempCx10;

u16 gnSensorState;

int gnSmpCnt;

s32 gZBuf[SNS_NUM_Z_SAMPLES];

s32 gnAltAvgM; 
s32 gnAltAvgCm;
s32 gnAltCm;
s32 gnCps;
s32 gnSum_t;
s32 gnSum_t2;
s32 gnDen;
s16 gnAccel=0;

s16 gnBeepPeriodTicks = 0;
s16 gnBeepTransition = 0;
s32 gnBeepCps = 0;
s32 gnVarioCps = 0;
s32 gnFreqHz = 0;

void audio_Config(void);
void audio_Beep(s32 nCps);
void audio_SetFreq(s32 nFreqHz);

void sns_TriggerPressureSample(void);
void sns_TriggerTemperatureSample(void);
void sns_DeltaCmPerSec(void);
void sns_Config(void);
void sns_InitData(void);
void sns_InitWindowBuffers(void);
void sns_ReadCoeffs(void);
void sns_Reset(void);
void sns_AcquireAveragedSample(int nSamples);
void sns_SampleContinuous(void);
s32  sns_Pa2Cm(s32 pa);
void sns_AverageAltitude(void);
void sns_Reset(void);
void sns_SendCmd(u32 cmd, u16 nbits);
u16  sns_ReadWord(void);
void sns_ConvertPressureTemperature(void);

void adc_batVoltage(void);
void mcu_Config(void);

#ifdef DEBUG
void uart_Print(char *szBuf);
void uart_Config(void);
void delay(unsigned int d);

char gszBuf[80];
#endif



void audio_SetFreq(s32 nFreqHz) {
   TACTL = TASSEL_2 + MC_0;  // stop timer A
   if (nFreqHz) {
      CCR0 = (u16)(FOSC_HZ/nFreqHz);
      CCR1 = (CCR0 >> 1); // square wave
      CCR2 = CCR1;
      }
   else { // dc 0V
      CCR0 = 0;
      CCR1 = 0;
      CCR2 = 0;
      }
   
   TACTL = TASSEL_2 + MC_1;  // start timer A
   }





//  +---+---+---+ BeepPeriod +----+----+----+----+----+----+----+
//                                     +----+--BeepTransition---+
//
//  Each tick is one pressure value sampling cycle  (approximately 1/11 seconds).
//  A counter is initialized to the value BeepPeriod at the start of the tone generation. 
//  Each tick, the counter is decremented. When it reaches the value BeepTransition, the tone
//  is shut off. The whole cycle repeats when the counter reaches zero.

// for acceleration 
const s16 gnBeepATbl[4][2] = {
{5,1}, 
{4,1},
{3,1}, 
{2,1} 
};

// for speed
const s16 gnBeepSTbl[20][2] = {
{5,1}, 
{5,1},
{4,1}, 
{4,1}, 
{3,1}, 
{3,1}, 
{3,1}, 
{3,1}, 
{2,1}, 
{2,1}, 
{2,1},
{2,1}, 
{2,1}, 
{2,1}, 
{2,1}, 
{2,1}, 
{2,1}, 
{2,1}, 
{2,1},
{2,1}  
};


void sns_LPFAccel(void)   {
   	s16 tmp;
	static s16 nAccelx32 = 0;
   	tmp = nAccelx32*KLPF_ACCEL + (gnAccel<<5)*(32-KLPF_ACCEL);
   	nAccelx32 = (tmp >= 0 ? ((tmp + 16)>>5) : ((tmp - 16)>>5));
   	gnAccel = (nAccelx32>>5);
   	} 


void audio_Beep(s32 nCps) {
   static s32 nFreqHz;
   static s32 nOldCps = 0;
   int nRange;

   gnAccel = (s16)(nCps - nOldCps);
   sns_LPFAccel();
   nOldCps = nCps;

   if ((gnBeepPeriodTicks <= 0) || (ABS(nCps-gnVarioCps) > BEEP_DISCR_THRESHOLD) || ((nCps > 0) && (gnVarioCps <= 0)) ) {
      gnVarioCps = nCps;
      // if sinking much faster than glider sink rate, generate continuous tone alarm
      if (gnVarioCps < SINKTONE_CPS) {
         gnBeepCps = gnVarioCps;	
         nFreqHz = VARIO_SINK_FREQHZ;
         audio_SetFreq(nFreqHz);
         gnBeepPeriodTicks = 0;
         gnBeepTransition  = 0;
         }
      //if climbing, generate beep
      else {
         gnBeepCps = gnVarioCps - CLIMBTONE_CPS;
         if (gnBeepCps > 0) {
            if (gnBeepCps > VARIO_MAX_CPS) {
		gnBeepCps = VARIO_MAX_CPS;
		}
	    if (OPT_ACCEL) { 
               nRange = ABS(gnAccel)/5;
               if (nRange > 3) nRange = 3;
               gnBeepPeriodTicks = gnBeepATbl[nRange][0];
               gnBeepTransition = gnBeepATbl[nRange][1];
	       }
	    else {
               nRange = gnBeepCps/50;
               if (nRange > 19) nRange = 19;
               gnBeepPeriodTicks = gnBeepSTbl[nRange][0];
               gnBeepTransition = gnBeepSTbl[nRange][1];
               }
           if (gnBeepCps > VARIO_XOVER_CPS) {
               nFreqHz = VARIO_XOVER_FREQHZ + ((gnBeepCps - VARIO_XOVER_CPS)*(VARIO_MAX_FREQHZ - VARIO_XOVER_FREQHZ))/(VARIO_MAX_CPS - VARIO_XOVER_CPS);
               }
           else {
               nFreqHz = VARIO_MIN_FREQHZ + (gnBeepCps*(VARIO_XOVER_FREQHZ - VARIO_MIN_FREQHZ))/VARIO_XOVER_CPS;
               }
           audio_SetFreq(nFreqHz);
           }
      // if  within normal glider sink rate, be quiet
      else{
           gnBeepPeriodTicks = 0;
           gnBeepTransition  = 0;
           nFreqHz = 0;
           audio_SetFreq(nFreqHz);
           }
        }
  }
 else{ 
      gnBeepPeriodTicks--;
      if (gnBeepPeriodTicks <= gnBeepTransition){ // shut off tone
         nFreqHz = 0;
         audio_SetFreq(nFreqHz);
         }
      else 
      if (OPT_CHIRP) {
	 nFreqHz += 10;
         audio_SetFreq(nFreqHz);
	 }
      }
   }
   
   
void mcu_Config (void) {
   WDTCTL = WDTPW + WDTHOLD;  // disable watchdog
   
   BCSCTL1 &= 0xF8; 
   BCSCTL1 |= 0x02;  // RSEL = 2, DCO ~ 278kHz
   
   P1SEL = 0x0C;     // P1.2  = TA1, P1.3 = TA2
   P1DIR = 0x0D;     // P1.0, P1.2, P1.3 are outputs
   P1OUT  = 0;               
   
   CCTL1 = OUTMOD_7; // CCR1 reset/set mode  
   CCTL2 = OUTMOD_3; // CCR2 set/reset mode   (inverted phase of CCTL1)

   P3DIR = 0x80;     // P3.7 is output (SCLK)
   P3OUT = 0;

   P4SEL = 0x04;    // P4.2 is TB2  (MS5534B MCLK)
   P4DIR = 0x06;    // P4.0 is input (DOUT), 4.1 (DIN) and 4.2 (MS5534B MCLK) are outputs 
   P4OUT = 0;
   
#ifdef DEBUG_CLKS   
   P5DIR = 0x30;    // P5.4, 5.5 test outputs used for checking system and peripheral clock freq
   P5SEL = 0x30;    // P5.4 = MCLK, 5.5 = SMCLK
#endif

   TBCCR0 =  (u16)(FOSC_HZ/34000L);  // TMRB 34kHz square wave sampling clock for MS5534B (max 35kHz)
   TBCCTL2 = OUTMOD_7;                // use set/reset mode
   TBCCR2 = TBCCR0 >> 1;              // square wave
   TBCTL = TBSSEL_2 + MC_1;           // start tmrB using SMCLK in  up mode

#ifdef DEBUG
   uart_Config();
#endif   
   }
   
      
void sns_Config(void) {   
   sns_Reset();
   sns_ReadCoeffs();
   
   // These parameters used in climb/sink rate calculation need to be computed only once.
   gnSum_t = -((s32)SNS_NUM_Z_SAMPLES * (SNS_NUM_Z_SAMPLES-1))/2;
   gnSum_t2 = ((s32)SNS_NUM_Z_SAMPLES * (SNS_NUM_Z_SAMPLES-1)*(2*SNS_NUM_Z_SAMPLES-1))/6;
   gnDen = ((s32)SNS_NUM_Z_SAMPLES*gnSum_t2) - (gnSum_t*gnSum_t);
   }


void sns_InitData(void) {
   sns_InitWindowBuffers();
   gnCps = 0;
   gnSmpCnt = SNS_NUM_Z_SAMPLES-1;

#ifdef DEBUG
   sprintf(gszBuf, "Alt %ldm", gnAltAvgM);
   uart_Print(gszBuf);
#endif
      
   sns_TriggerTemperatureSample();
   gnSensorState = SENSOR_READ_TEMPERATURE;
   }


// Each pressure or temperature conversion takes ~34mS
// At ~250kHz MCLK, we use max 28mS for the computation, so still have some headroom

void sns_SampleContinuous(void)  {  
   if (DOUT) return;  // conversion not yet complete
   if (gnSensorState == SENSOR_READ_PRESSURE)  {
      gwD1 = sns_ReadWord(); // read pressure
      sns_TriggerTemperatureSample();
      P1OUT |= B0;
      sns_ConvertPressureTemperature();      
      gZBuf[gnSmpCnt] = gnAltCm = sns_Pa2Cm(gP*10); // ms5534 has 0.1mbar (10pa) resolution, LUT has 1 pa input resolution. 
      gnTempCx10 = gT;
      sns_AverageAltitude();
      sns_DeltaCmPerSec();
      audio_Beep(gnCps);
      gnSmpCnt++;
      if (gnSmpCnt == SNS_NUM_Z_SAMPLES)  {
         gnSmpCnt = 0;
         }
      gnSensorState = SENSOR_READ_TEMPERATURE;
      P1OUT &= ~B0; // when debugging, check P1.0 waveform, high time should be less than 34 mS
      }
   else // gnSensorState == SESOR_READ_TEMPERATURE
      {
      gwD2 = sns_ReadWord(); // read temperature
      sns_TriggerPressureSample();
      gnSensorState = SENSOR_READ_PRESSURE;
      }
   }



void sns_InitWindowBuffers(void) {
   int n;
   sns_AcquireAveragedSample(4);
   n = SNS_NUM_Z_SAMPLES;
   while(n--) {
      gZBuf[n] = gnAltCm;
      }
   }
   
   
u16 sns_ReadWord(void)  {
   u16 w;
   u16 clk = 16;
   w = 0;
   while (clk--)  {
      SCLK_1();
      SCLK_0();
      w |=  (DOUT << clk);
      }
   SCLK_1();
   SCLK_0();
   return w;
   }


// 1010101010101010 + 00000
void sns_Reset(void) {
   sns_SendCmd(0x155540,21);
   }

// send command MS bit first
void sns_SendCmd(u32 cmd, u16 nbits)  {
   while (nbits--)  {
      if (cmd & (u32)(1<<nbits)) {
        DIN_1(); 
        }
     else {
        DIN_0();
        }
      SCLK_1();
      SCLK_0();
      }
   }

// 111 + 6bit coeff addr + 000 + 1clk(send0) 
u16 sns_ReadCoeff(u8 addr) {
   u32 cmd = (u32)0x1C00 | (((u32)addr)<<4);
   sns_SendCmd(cmd,13);
   return sns_ReadWord();
   }

// 111 + 1010 + 000 + 2clks(send 0)
void sns_TriggerPressureSample(void) {
   sns_Reset();
   sns_SendCmd(0xF40,12);
   }

// 111 + 1001 + 000 + 2clks(send 0)
void sns_TriggerTemperatureSample(void) {
   sns_Reset();
   sns_SendCmd(0xF20,12);
   }


void sns_ReadCoeffs(void) {
   u16 wa,wb;
   wa = sns_ReadCoeff(0x15);
   wb = sns_ReadCoeff(0x16);
   gwC1 = (u16)((wa >> 1) & (u16)0x7FFF);
   gwC5 = (u16)(((wa & 0x1)<<10) | ((wb >> 6) & (u16)0x3FF));
   gwC6 = (u16)(wb & 0x3F);

   wa = sns_ReadCoeff(0x19);
   wb = sns_ReadCoeff(0x1A);
   gwC4 = (u16)((wa>>6) & 0x3FF);
   gwC2 = (u16)(((wa & 0x3F) << 6) | (wb & 0x3F));
   gwC3 =  (u16)((wb >> 6) & 0x3FF);
#ifdef DEBUG
   sprintf(gszBuf,"%u %u %u %u %u %u\r\n",gwC1,gwC2,gwC3,gwC4,gwC5,gwC6);
   uart_Print(gszBuf);
#endif
   }


void sns_ConvertPressureTemperature(void) {
    s32 UT1,dT,OFF,SENS,X,TEMP,P2,T2;
    UT1  = (gwC5 <<3) + 20224;
    dT   = (s32)gwD2  - UT1;
    TEMP = 200 + ((dT*(gwC6 +50))>>10);
    OFF  = (gwC2 <<2) + (((gwC4 -512)*dT)>>12);
    SENS = gwC1  + ((gwC3 *dT)>>10)  + 24576;
    X    = ((SENS* ((s32)gwD1  - 7168))>>14)  - OFF;
    gP  = ((X*10)>>5) + 2500;
    if (TEMP < 200) {
        T2 = (11*(gwC6 +24)*(200-TEMP)*(200-TEMP))>>20;
        P2 = (3*T2*(gP -3500))>>14;
        }
    else  {
        T2 = 0;
        P2 = 0;
        }
    TEMP = TEMP-T2;
    gP  = gP -P2;
    gT  = (s16)TEMP;  
    }


void sns_AcquireAveragedSample(int nSamples) {
   s32 pa,pAccum,n;
   pAccum = 0;
   n = nSamples;
   while (n--) {
      sns_TriggerTemperatureSample();
      while(DOUT);
      gwD2 = sns_ReadWord(); // read temperature
      sns_TriggerPressureSample();
      while (DOUT);
      gwD1 = sns_ReadWord(); // read pressure
      sns_ConvertPressureTemperature();
      pAccum += gP;
      }
   pa = pAccum/nSamples;
   gnAltCm = sns_Pa2Cm(pa*10);
   gnAltAvgM = (gnAltCm >= 0 ? (gnAltCm+50L)/100L : (gnAltCm-50L)/100L);
#ifdef DEBUG
   sprintf(gszBuf,"Averaged Altitude : %ldm\r\n",gnAltAvgM);
   uart_Print(gszBuf);
#endif
   }


void sns_AverageAltitude(void) {
   int cnt;
   s32 avg;
   cnt = SNS_NUM_Z_SAMPLES;
   avg = 0;
   while (cnt--) {
       avg += gZBuf[cnt];
       }
   avg /= SNS_NUM_Z_SAMPLES;
   gnAltAvgCm = avg;
   gnAltAvgM = ((avg >= 0) ? (avg+50L)/100L : (avg-50L)/100L); 
   return; 
   }

void sns_DeltaCmPerSec(void) {
   int n,trel;
   s32 z,sum_zt;
   sum_zt = 0;
   n = SNS_NUM_Z_SAMPLES;
   while (n--)  {
      z = gZBuf[n] - gnAltAvgCm;
      trel = n - gnSmpCnt;    // time origin is the current sample in z buffer
      if (n > gnSmpCnt) {
         trel -= SNS_NUM_Z_SAMPLES;
         }
      sum_zt += (z*(s32)trel);
      }
   gnCps = (SNS_NUM_Z_SAMPLES*sum_zt*SENSOR_SAMPLES_PER_SEC)/gnDen;
   CLAMP(gnCps,-32768,32767);
   return;
   }


// Lookup table converting pressure in Pa to altitude in cm.
// Each LUT entry is the altitude in cm corresponding to an implicit
// pressure value, calculated as [PA_INIT - 1024*index] in Pa.
// The table is calculated for a nominal sea-level pressure  = 101325 Pa.


#define PZLUT_ENTRIES   77
#define PA_INIT         104908L
#define PA_DELTA        1024L

const s32 gPZTbl[PZLUT_ENTRIES] = {
-29408,
-21087,
-12700,
-4244,
4279,
12874,
21541,
30281,
39095,
47986,
56953,
66000,
75126,
84335,
93628,
103006,
112472,
122026,
131672,
141410,
151244,
161174,
171204,
181335,
191570,
201911,
212361,
222922,
233597,
244388,
255300,
266334,
277494,
288782,
300204,
311761,
323457,
335297,
347285,
359424,
371719,
384174,
396795,
409586,
422552,
435700,
449033,
462560,
476285,
490216,
504360,
518724,
533316,
548144,
563216,
578543,
594134,
609999,
626149,
642595,
659352,
676431,
693847,
711615,
729752,
748275,
767202,
786555,
806356,
826627,
847395,
868688,
890537,
912974,
936037,
959766,
984206
};




s32 sns_Pa2Cm(s32 pa)  {
   	s32 inx,pa1,z1,z2,z;
   	if (pa > PA_INIT) {
      	z = gPZTbl[0];
      	}
   	else {
      	inx = (PA_INIT - pa)>>10;      
      	if (inx >= PZLUT_ENTRIES-1) {
         	z = gPZTbl[PZLUT_ENTRIES-1];
         	}
      	else {
         	pa1 = PA_INIT - (inx<<10);
         	z1 = gPZTbl[inx];
         	z2 = gPZTbl[inx+1];
         	z = z1 + (((pa1-pa)*(z2-z1))>>10);
         	}
      	}
   	return z;
   	}

#ifdef DEBUG

void uart_Config(void) {
   P3SEL = 0x30;            // P3.4, 3.5 = USART0 TXD/RXD
   ME1 |= UTXE0;            // Enable USART0 TXD
   UCTL0 |= CHAR;           // 8-bit character
   UTCTL0 |= SSEL_2;        // UCLK = MCLK
   UBR00 = (u8)(FOSC_HZ/9600L);            
   UBR10 = 0;             
   UMCTL0 = 0x00;           // no modulation
   UCTL0 &= ~SWRST;         // Initialize USART state machine
   }

void uart_Print(char *szBuf) {
   
    while (*szBuf) {
       while (!(IFG1 & UTXIFG0));
       TXBUF0 = *szBuf++;    
      
       }

   }


#endif

void delay(u16 d) {
   while (d--) {
        nop();
        nop();
        nop();
        nop();
        nop();
        nop();
        nop();
        nop();
        nop();
        nop();
        nop();
        nop();
        nop();
        }
    }

void adc_BatVoltage(void) {
    u16 batVoltage;
    int n;
    P6SEL |= 0x01;                            // Enable A/D channel A0
    ADC12CTL0 = ADC12ON+SHT0_2;               // Turn on and set up ADC12
    ADC12CTL1 = SHP;                          // Use sampling timer
    ADC12CTL0 |= ENC;                         // Enable conversions

    ADC12CTL0 |= ADC12SC;                   // throw away results of first conversion
    while ((ADC12IFG & BIT0)==0);          // 
    
    n = 8;
    batVoltage = 0;
    while (n--) {
        ADC12CTL0 |= ADC12SC;                   // Start conversion
        while ((ADC12IFG & BIT0)==0);          // wait for end of conversion
        batVoltage += ADC12MEM0;
#ifdef DEBUG        
        sprintf(gszBuf,"Sample %d\r\n", ADC12MEM0); 
        uart_Print(gszBuf);
#endif        
        }
    batVoltage >>= 3;   // average 8 readings
    // Vcc = 2.5V is reference, (25/4095) = (5/819)
    batVoltage *= 5;
    batVoltage /= 819;
#ifdef DEBUG
    sprintf(gszBuf,"ADCV %d.%dV\r\n",batVoltage/10,batVoltage%10); 
    uart_Print(gszBuf);
#endif
    batVoltage *= 27;  // resistor divider is 0.373, so actual voltage is  v / 0.373 = v * 2.7
    batVoltage += 50;  // round off to nearest volt
    batVoltage /= 100;
    while (batVoltage--) {
        audio_SetFreq(1500);
        delay(0xfff);
        audio_SetFreq(0);
        delay(0xfff);
        }
        
    ADC12CTL0 = 0;  // turn off ADC to conserve power
    P6SEL &= 0xFE;
    }



int main(void) {
    
   mcu_Config();
   
#ifdef DEBUG
   delay(0xffff);
#endif

   delay(0xff);
   adc_BatVoltage();      
   sns_Config();
   sns_InitData();

   while (1) {                       
      sns_SampleContinuous();
      }
   return 0;
   }

